#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"

import sys
import os
import errno
import re
import xml.etree.ElementTree as ET
from xml.parsers.expat import ExpatError

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import HCPClient
import HDIConfig
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception
import Utility

sys.path.append(os.path.join(BASEDIR, "bin"))
import Exceptions

# Constants
hdiconf_dir_on_hcp = "/rest/foscripts/hdiconf/"
synced_dir = "synced"
backup_disabled_file = 'backup_disabled'

# Functions

def get_synced_dir_path(config_dir):
    synced_dir_path = os.path.join(config_dir, synced_dir)
    return synced_dir_path

def get_synced_fsinfo(fsname, ipaddr, config_dir):
    log_info("Getting synced file system information of %s on %s ..." % (fsname, ipaddr))

    synced_dir_path = get_synced_dir_path(config_dir)

    # Get synced file system information
    fsinfo_file_path = os.path.join(synced_dir_path, make_fsinfo_file_name(fsname, ipaddr))
    if not os.path.exists(fsinfo_file_path):
        # Previous file system information is not found
        return None

    try:
        fsinfo = HDIConfig.parse_filesystem(xml_file=fsinfo_file_path)
    except HDIConfig.FormatError, e:
        log_error("A wrong information was detected in a synced file system information. file=%s, error=%s" % (fsinfo_file_path, str(e)))
        # Treat it as no synced information found
        return None

    return fsinfo

def get_synced_fslist(config_dir):
    """ Get a list of file systems that were synced on this node.

    Args:
        config_dir (str): Directory path where HDI configuration archive file
            is stored.
    Returns:
        dict: Dictionary (key is IP address of a node, value is list of file
            system name that were synced from)
    """
    log_info("Getting a list of file systems synced on this node ...")

    synced_dir_path = get_synced_dir_path(config_dir)
    try:
        file_list = os.listdir(synced_dir_path)
    except EnvironmentError, e:
        if e.errno == errno.ENOENT:
            # synced directory may not exist. Just returns as "nothing found".
            return {}
        else:
            log_error("Failed to get a list of files on %s. error=%s" % (synced_dir_path, str(e)))
            raise

    synced_fs_dict = {}
    for file_name in file_list:
        info = get_node_and_fsname_from_fsinfo_file(file_name)
        if info is not None:
            ipaddr, fsname = info
            if ipaddr not in synced_fs_dict:
                synced_fs_dict[ipaddr] = []
            synced_fs_dict[ipaddr].append(fsname)

    return synced_fs_dict

def save_fsinfo(fsinfo, ipaddr, config_dir):
    log_info("Saving file system information of %s on %s ..." % (fsinfo.fileSystemName, ipaddr))

    synced_dir_path = get_synced_dir_path(config_dir)
    try:
        os.mkdir(synced_dir_path)
    except EnvironmentError, e:
        if e.errno != errno.EEXIST:
            log_error("Failed to create a directory %s. error=%s" % (synced_dir_path, str(e)))
            raise

    fsinfo_file_path = os.path.join(synced_dir_path, make_fsinfo_file_name(fsinfo.fileSystemName, ipaddr))
    try:
        with open(fsinfo_file_path, 'w') as f:
            f.write(fsinfo.xml)
    except EnvironmentError, e:
        log_error("Failed to save file system information to %s. error=%s" % (fsinfo_file_path, str(e)))
        raise

def delete_fsinfo(fsname, ipaddr, config_dir):
    log_info("Deleting file system information of %s on %s ..." % (fsname, ipaddr))

    synced_dir_path = get_synced_dir_path(config_dir)
    fsinfo_file_path = os.path.join(synced_dir_path, make_fsinfo_file_name(fsname, ipaddr))
    log_info("Deleting %s ..." % fsinfo_file_path)
    Utility.remove_file(fsinfo_file_path)

def get_node_and_fsname_from_fsinfo_file(file_name):
    m = re.match(r"fsinfo_([0-9\.]+)_([0-9a-zA-Z_]+).xml", file_name)
    if m is None:
        return None
    return (m.group(1), m.group(2))

def make_fsinfo_file_name(fsname, ipaddr):
    return "fsinfo_%s_%s.xml" % (ipaddr, fsname)

def get_synced_policy(fsname, ipaddr, config_dir):
    """ Get storage policy of a file system that was applied at the last
    synchronization.

    Args:
        fsname (str): File system name.
        ipaddr (str): IP address of the node that the file system was comming
            from.
    Returns:
        str: Storage policy string of the file system that was applied at the
            last synchronization.
    """
    log_info("Getting synced storage policy of %s on %s ..." % (fsname, ipaddr))

    synced_dir_path = get_synced_dir_path(config_dir)
    policy_file_path = os.path.join(synced_dir_path, make_policy_file_name(fsname, ipaddr))
    if not os.path.exists(policy_file_path):
        # Not found
        return None

    try:
        with open(policy_file_path, 'r') as f:
            policy_string = f.read()
    except EnvironmentError, e:
        log_error("Failed to read a file %s. error=%s" % (policy_file_path, str(e)))
        # Treat it as no synced information found
        return None

    return policy_string.strip()

def save_policy(policy_string, fsname, ipaddr, config_dir):
    log_info("Saving storage policy information of %s on %s ..." % (fsname, ipaddr))

    synced_dir_path = get_synced_dir_path(config_dir)
    try:
        os.mkdir(synced_dir_path)
    except EnvironmentError, e:
        if e.errno != errno.EEXIST:
            log_error("Failed to create a directory %s. error=%s" % (synced_dir_path, str(e)))
            raise

    policy_file_path = os.path.join(synced_dir_path, make_policy_file_name(fsname, ipaddr))
    try:
        with open(policy_file_path, 'w') as f:
            f.write(policy_string)
    except EnvironmentError, e:
        log_error("Failed to save storage policy information of %s to %s. error=%s" % (fsname, policy_file_path, str(e)))
        raise

def delete_policy(fsname, ipaddr, config_dir):
    log_info("Deleting storage policy information of %s on %s ..." % (fsname, ipaddr))

    synced_dir_path = get_synced_dir_path(config_dir)
    policy_file_path = os.path.join(synced_dir_path, make_policy_file_name(fsname, ipaddr))
    log_info("Deleting %s ..." % policy_file_path)
    Utility.remove_file(policy_file_path)

def is_policy_of(file_name, ipaddr, fsname=None):
    if fsname is not None:
        m = re.match(r"policy_%s_%s.txt" % (ipaddr, fsname), file_name)
    else:
        m = re.match(r"policy_%s_[0-9a-zA-Z_]+.txt" % ipaddr, file_name)
    if m is None:
        return False
    else:
        return True

def make_policy_file_name(fsname, ipaddr):
    return "policy_%s_%s.txt" % (ipaddr, fsname)

def make_configs_archive_name(ipaddr, hostname, system_uuid):
    return "hdiconf_%s_%s_%s.tar.gz" % (ipaddr, hostname, system_uuid)

def is_configs_archive_of(archive_name, ipaddr):
    """ Determines if an archive file is an archive for a specific node.

    This function determines if an archive file is an archive for a specific
    node, by checking its file name.
    Returns True if the archive file is an archive file for the node. Otherwise
    returns False.

    Args:
        archive_name (str): Archive file name to be checked. Only name, not a path.
        ipaddr (str): IP address of the node.
    Returns:
        bool: True if the archive file is an archive of the node. Otherwise False.
    """

    m = re.match(r"hdiconf_%s_[0-9a-zA-Z-]+_[0-9a-fA-F-]+\.tar\.gz" % ipaddr, archive_name)
    if m is None:
        return False
    else:
        return True

def is_configs_archive(archive_name):
    """ Determines if the specified name is a configuration archive file name.

    This function is similar to is_configs_archive_of() but this function is
    independent from IP address.

    Args:
        archive_name (str): Archive file name to be checked. Only name, not a path.
    Returns:
        bool: True if the archive file is a configuration archive. Otherwise False.
    """
    m = re.match(r"hdiconf_[0-9.]+_[0-9a-zA-Z-]+_[0-9a-fA-F-]+\.tar\.gz", archive_name)
    if m is None:
        return False
    else:
        return True

def get_list_of_archive_files_on_HCP(hcp_info, config):
    hcp_client = HCPClient.HCPClient(hcp_info, "system-backup-data")
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)
    try:
        ret = hcp_client.get(hdiconf_dir_on_hcp)
    except HCPClient.ConnectionError, e:
        log_error("Failed to communicate with HCP when getting a list of archive files. error=%s" % str(e))
        raise Exceptions.ConnectionError("Connection error with HCP (%s)" % e.reason, e.code)

    if not Utility.is_expected_HTTP_status(ret, (200,)):
        raise Exceptions.OperationError("Failed to get a list of HDI configuration archive files stored on HCP. error=%s" % ret['message'], ret['code'])

    xml = ret['body'].strip()

    # Find an archive file for the specified node
    try:
        root = ET.fromstring(xml)
    except ExpatError, e:
        log_error("Invalid XML syntax in the XML returned by HCP. code=%s, line=%s, offset=%s, xml=%s" % (e.code, e.lineno, e.offset, xml))
        raise Exceptions.OperationError("Unexpected XML contents returned from HCP.", -1)

    entries = root.findall('entry')
    return [entry.attrib['urlName'] for entry in entries]

def delete_configs_on_HCP(ipaddr, hcp_info, config):
    """ Delete configuration archive file of a specified node on HCP.

    This function deletes all configuration files of a specified node on HCP.

    Args:
        ipaddr (str): IP address of a node which configuration files to be deleted.
        hcp_info (HCPClient.HCPInfo): HCP information.
    Returns:
        None
    """

    log_info("Delete configuration archive files of %s from HCP." % ipaddr)

    # Get a list of archive files stored on HCP
    archive_list_on_HCP = get_list_of_archive_files_on_HCP(hcp_info, config)
    log_info("Archive files on HCP: %s" % archive_list_on_HCP)

    # Find an archive file for the specified node
    list_to_be_deleted = []
    for archive_name in archive_list_on_HCP:
        if is_configs_archive_of(archive_name, ipaddr):
            list_to_be_deleted.append(archive_name)

    if len(list_to_be_deleted) == 0:
        log_info("No configuration archive files for %s found." % ipaddr)
        return

    # Delete the files
    hcp_client = HCPClient.HCPClient(hcp_info, "system-backup-data")
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)
    for archive_name in list_to_be_deleted:
        log_info("Deleting %s on HCP ..." % archive_name)
        archive_path_on_hcp = os.path.join(hdiconf_dir_on_hcp, archive_name)
        try:
            ret = hcp_client.delete(archive_path_on_hcp)
        except HCPClient.ConnectionError, e:
            log_error("Failed to communicate with HCP. error=%s" % str(e))
            raise Exceptions.ConnectionError("Connection error with HCP (%s)" % e.reason, e.code)

        if not Utility.is_expected_HTTP_status(ret, (200, 404)):
            raise Exceptions.OperationError("Failed to download an HDI configuration archive file to HCP. error=%s" % ret['message'], ret['code'])

def download_configs(ipaddr, config_dir, hcp_info, config):
    """ Download configuration archive from HCP.

    Download configuration archive of the specified node from HCP, and save
    it to the specified directory.
    If there is already
    Args:
        hcp_info (HCPClient.HCPInfo): HCP information.
        ipaddr (str): IP address of the node of the configuration archive
            to be downloaded.
        config_dir (str): Base directory path where the configuration archive
            is saved.
    Returns:
        str: A path of the downloaded configuration archive file.
    """

    log_info("Download an archive file from HCP. node=%s, path=%s" % (ipaddr, hdiconf_dir_on_hcp))

    # Get a list of archive files stored on HCP
    archive_list_on_HCP = get_list_of_archive_files_on_HCP(hcp_info, config)
    log_info("Archive files on HCP: %s" % archive_list_on_HCP)

    # Find an archive file for the specified node
    archive_name = None
    for archive_name in archive_list_on_HCP:
        if is_configs_archive_of(archive_name, ipaddr):
            break
    else:
        log_warning("Archive file of %s does not exist on HCP." % ipaddr)
        archive_name = None # Just to be explicit
        return None

    hcp_client = HCPClient.HCPClient(hcp_info, "system-backup-data")
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)

    archive_path_on_hcp = os.path.join(hdiconf_dir_on_hcp, archive_name)
    archive_path_on_hdi = os.path.join(config_dir, archive_name)

    # Delete an existing archive file
    existing_archive_path = get_existing_archive_path(ipaddr, config_dir)

    if existing_archive_path is not None:
        log_info("Deleting an existing archive file %s ..." % existing_archive_path)
        Utility.remove_file(existing_archive_path)

    # Download the archive file
    try:
        ret = hcp_client.get(archive_path_on_hcp, to_file=archive_path_on_hdi)
    except HCPClient.ConnectionError, e:
        log_error("Failed to communicate with HCP when downloading an archive file. error=%s" % str(e))
        raise Exceptions.ConnectionError("Connection error with HCP (%s)" % e.reason, e.code)

    if not Utility.is_expected_HTTP_status(ret, (200,)):
        raise Exceptions.OperationError("Failed to download an HDI configuration archive file to HCP. error=%s" % ret['message'], ret['code'])

    return archive_path_on_hdi

def get_existing_archive_path(ipaddr, config_dir):
    archive_list = Utility.get_list_of_files(config_dir)
    for archive_name in archive_list:
        if is_configs_archive_of(archive_name, ipaddr):
            existing_archive_path = os.path.join(config_dir, archive_name)
            break
    else:
        existing_archive_path = None

    return existing_archive_path

def get_list_of_archive_files_on_HDI(config_dir):
    archive_path_list = []
    for name in Utility.get_list_of_files(config_dir):
        if is_configs_archive(name):
            archive_path_list.append(os.path.join(config_dir, name))

    return archive_path_list

def disable_backup(config_dir):
    backup_disabled_file_path = os.path.join(config_dir, backup_disabled_file)
    log_info("Disabling backup_configs execution ...")
    try:
        with open(backup_disabled_file_path, 'w'):
            # Just create an empty file
            pass
    except EnvironmentError, e:
        log_error("Failed to create a flag file %s. error=%s" % (backup_disabled_file_path, str(e)))
        raise

def enable_backup(config_dir):
    backup_disabled_file_path = os.path.join(config_dir, backup_disabled_file)
    log_info("Enabling backup_configs execution ...")
    Utility.remove_file(backup_disabled_file_path)

def is_backup_disabled(config_dir):
    backup_disabled_file_path = os.path.join(config_dir, backup_disabled_file)
    return os.path.exists(backup_disabled_file_path)
