#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "check_resources"

import sys
import os
import getopt

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_warning, print_status_error
import HDIClient
import HDIConfig
import HDIOperation
import HCPClient
import HCPMQE
import FailoverConfig

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import HDIConfigArchive
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import setup_common
import getinfo_common
import backup_sync_common

def usage():
    print "Usage:"
    print "    sudo maintexec %s" % PROGRAM
    print "    sudo maintexec %s -h|--help" % PROGRAM

def check_resources(myself, config, nobackup=False, nocheckMQE=False):
    """ Check resource conflicts and shortage on this node.

    This function checks if there are any name conflicts of failover resources
    (file system name, share name, migration task name) which will be failed
    over to this node. Also this function checks if there is any storage shortage
    if all file systems on all primary nodes paired with this node are failed
    over to this node.
    If there is any conflicts or shortage, this function returns True. Otherwise
    False.

    This function at first downloads configuration archives of all primary nodes
    that are paired with this node. If nobackup is False, this function executes
    backup_configs command on each primary node before downloading the archives.

    This function uses the same way to allocate storage for a file system with
    the actual failover when checking storage shortage.

    Args:
        myself (str): IP address of this node.
        config (Config.Config): Configuration of the scripts.
        nobackup (bool): If this is True, this function does not execute
            backup_configs on each primary node. Otherwise backup_configs is
            executed on each primary node before downloading configuration archives.
        nocheckMQE (bool): If this is True, this function does not check availablity
            of MQE on the HCP and availablity of Search and Indexing on the namespace.

    Returns:
        dict: Returns the result of the check.
            {
                'has_conflict': bool,
                'has_shortage': bool,
                'is_MQE_unavailable': bool
            }
    """
    # Get list of configuration group name
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Setup connection to localhost
    # SSH credential is not required here because this is connection to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn_local = HDIOperation.Connection("localhost", cred, config.retry_count, config.retry_interval, config.timeout)

    # Load HCP information
    print_status("Loading HCP configuration ...")
    try:
        hcpop = HDIOperation.HCPSetting(conn_local)
        hcp_info = hcpop.get()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HCP information on the node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get HCP configuration on the node. error=%s" % e.reason, e.code)

    # Load node information
    print_status("Getting node information ...")
    nodeop = HDIOperation.Node(conn_local)
    try:
        model = nodeop.get_model()
        othernode = nodeop.get_other_mng_ip()
        node_number = nodeop.get_number()
    except HDIOperation.OperationError, e:
        log_error("Failed to get node information on the node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get node information on the node. error=%s" % e.reason, e.code)
    log_info("model: %s, node_number: %s, other node: %s" % (HDIConfig.Model.tostr(model), node_number, othernode))

    # Get file system information that was already created on this node
    print_status("Getting file system information already created on this node ...")
    fslist_local = getinfo_common.get_local_file_systems(config)

    # Get HDI configuration of each primary node of my pairs
    print_status("There are %s configuration group(s) that this node belongs to." % len(confgroup_list))

    node_dict = {}
    archive_dict = {}
    for confgroup in sorted(confgroup_list):
        print_status("Getting resource information of primary nodes for configuration group %s." % confgroup)

        # Load pair configuration for this configuration group
        pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)

        # Setup SSHConfig for this configuration group
        ssh_config = Config.SSHConfig(config.sys_conf_dir, pairconf, confgroup)

        # Get primary nodes of myself
        my_pairs = pairconf.get_list(secondary=myself)
        primary_nodes = [pair.primary for pid, pair in my_pairs]
        print_status("Primary nodes: %s, configuration group: %s" % (primary_nodes, confgroup))

        # Get primary nodes of the other node if I am a node in cluster
        if model == HDIConfig.Model.CLUSTER:
            # Get IP addresses of the other node
            other_iplist = getinfo_common.get_remote_iplist(othernode, config, ssh_config)

            # Find the other node in the pair configuration
            other_ip = FailoverConfig.PairConfig.find_matched_ip(config.fo_conf_dir, other_iplist, confgroup)

            # Get a list of primary nodes paired with the other node
            other_pairs = pairconf.get_list(secondary=other_ip)
            other_primary_nodes = [pair.primary for pid, pair in other_pairs]
            print_status("Primary nodes of the other node: %s, configuration group: %s" % (other_primary_nodes, confgroup))
        else:
            other_primary_nodes = []

        if not nobackup:
            # Execute backup of HDI configuration on the primary nodes
            print_status("=== Execute backup on primary nodes ===")

            # Command
            cmd = "sudo maintexec %s" % os.path.join(BASEDIR, "backup_configs")
            op = HDIOperation.CommandOperation(cmd, display_stdout=True)

            # Execute on each primary node
            for node in primary_nodes + other_primary_nodes:
                print_status("Executing configuration backup on %s ..." % node)
                # Setup connection to the primary node
                cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(node), None)
                conn = HDIOperation.Connection(node, cred, config.retry_count, config.retry_interval, config.timeout)
                try:
                    ret, out, err = op.execute(conn)
                except HDIOperation.ConnectionError, e:
                    log_error("Failed to connect %s over SSH. code=%s, reason=%s" % (conn.host, e.code, e.reason))
                    raise Exceptions.ConnectionError(e.reason, e.code)

                if ret != 0:
                    log_error("Failed to backup configuration on %s. ret=%s" % (conn.host, ret))
                    raise Exceptions.OperationError("Failed to backup configuration on %s. ret=%s" % (conn.host, ret), ret)

        # Download HDI configuration from HCP
        print_status("=== Download configuration from HCP ===")
        for node in primary_nodes + other_primary_nodes:
            print_status("Downloading configuration of %s from HCP ..." % node)
            archive_path = backup_sync_common.download_configs(node, config.hdi_conf_dir, hcp_info, config)
            if archive_path is None:
                log_error("No archive file found for %s." % node)
                raise Exceptions.ConfigurationError("Configuration of %s has not been backed up." % node)

            archive = HDIConfigArchive.HDIConfigArchive()
            archive.load(archive_path)
            archive_dict[node] = archive

        node_dict[confgroup] = {'self': primary_nodes, 'other': other_primary_nodes}

    ### Check name conflicts
    has_conflict = False
    print_status("=== Name conflict check ===")

    # Display file systems
    print_status("File systems (unsupported file systems are excluded):")
    fslist_all = []
    for node in archive_dict:
        archive = archive_dict[node]
        fslist = [fsinfo for fsinfo in archive.fslist if fsinfo.is_supported()]
        fslist_all.extend(fslist)
        for fsinfo in fslist:
            print_status("    %s: %s" % (node, fsinfo.fileSystemName))

    # Check conflicts of file system name
    fsname_all = [fsinfo.fileSystemName for fsinfo in fslist_all]
    conflict_name_list = check_conflicts(fsname_all)
    if len(conflict_name_list) == 0:
        print_status("Result: No conflicts.")
    else:
        has_conflict = True
        print_status_warning("Result: CONFLICT EXISTS:")
        for name in conflict_name_list:
            print_status("    %s" % name)

    # Display CIFS share names
    print_status("CIFS shares:")
    cifslist_all = []
    for node in archive_dict:
        archive = archive_dict[node]
        cifslist = [cifsinfo for cifsinfo in archive.cifslist if cifsinfo.filesystem in fsname_all]
        cifslist_all.extend(cifslist)
        for cifsinfo in cifslist:
            print_status("    %s: %s" % (node, cifsinfo.cifsShareName))

    # Check conflicts of CIFS share name
    cifsname_all = [cifsinfo.cifsShareName for cifsinfo in cifslist_all]
    conflict_name_list = check_conflicts(cifsname_all)
    if len(conflict_name_list) == 0:
        print_status("Result: No conflicts.")
    else:
        has_conflict = True
        print_status_warning("Result: CONFLICT EXISTS:")
        for name in conflict_name_list:
            print_status("    %s" % name)

    # Display NFS share names
    print_status("NFS shares:")
    nfslist_all = []
    for node in archive_dict:
        archive = archive_dict[node]
        nfslist = [nfsinfo for nfsinfo in archive.nfslist if nfsinfo.filesystem in fsname_all]
        nfslist_all.extend(nfslist)
        for nfsinfo in nfslist:
            print_status("    %s: %s" % (node, nfsinfo.exportPoint))

    # Export path never conflicts as long as file system name does not conflict.
    print_status("Result: No conflicts. (Does not check NFS share conflicts - NFS shares never conflict as long as file systems do not conflict)")

    # Display migration task names
    print_status("Migration tasks:")
    tasklist_all = []
    for node in archive_dict:
        archive = archive_dict[node]
        tasklist = [taskinfo for taskinfo in archive.tasklist if taskinfo.fileSystemName in fsname_all]
        tasklist_all.extend(tasklist)
        for taskinfo in tasklist:
            print_status("    %s: %s on %s" % (node, taskinfo.task_name, taskinfo.fileSystemName))

    # Migration task never conflicts.
    # We can create multiple migration tasks which have the same name on a
    # different file system.
    print_status("Result: No conflicts. (Does not check migration task conflicts - task name conflict is allowed on a different file system)")

    ### Check storage conflicts and shortage
    has_shortage = False
    print_status("=== Storage resource shortage/conflict check ===")

    # Storage operation
    storage = HDIOperation.Storage(conn_local)

    # For each configuration group in dictionary order...
    for confgroup in sorted(confgroup_list):
        print_status("Checking file systems for configuration group %s ..." % confgroup)

        # Check file systems for node 0 first, then node 1
        if node_number == 0:
            nodes_for_node0 = node_dict[confgroup]['self']
            nodes_for_node1 = node_dict[confgroup]['other']
        else:
            nodes_for_node0 = node_dict[confgroup]['other']
            nodes_for_node1 = node_dict[confgroup]['self']

        for nodes in (nodes_for_node0, nodes_for_node1):
            fs_policy_list_for_node = []

            # Gather all file system information for node 0 (or node 1) in a list
            for node in nodes:
                archive = archive_dict[node]
                for fsinfo in [fsinfo for fsinfo in archive.fslist if fsinfo.is_supported()]:
                    if node in node_dict[confgroup]['self']:
                        target = myself
                    elif node in node_dict[confgroup]['other'] and 'other_ip' in locals():
                        # other_ip should be defined if node_dict[confgroup]['other'] is not empty
                        # but checks its existence just in case to avoid NameError.
                        target = other_ip
                    else:
                        target = myself
                    policy = get_storage_policy(fsinfo.fileSystemName, target, node, confgroup, config)
                    if policy is None:
                        print_status_error("No policy found for %s. Skip this file system ..." % fsinfo.fileSystemName)
                        continue
                    fs_policy_list_for_node.append((fsinfo, policy))

            # Sort the list in dictionary order of file system name
            fs_policy_list_for_node.sort(key=lambda e: e[0].fileSystemName)

            # Pick up LUs (cluster) or allocate capacity (single) for each file system
            for fsinfo, policy in fs_policy_list_for_node:
                if model == HDIConfig.Model.CLUSTER:
                    # If the file system was already created, display the information
                    if fsinfo.fileSystemName in [e.fileSystemName for e in fslist_local]:
                        fsinfo_local = [e for e in fslist_local if e.fileSystemName == fsinfo.fileSystemName][0]
                        fs_capacity = fsinfo_local.fileSystemSizeInGigaByte
                        # TODO: Need all LU information to get work space capacity for an existing file system
                        ws_capacity = "(not displayed for an existing file system)"
                        fs_lunamelist = [lu.deviceFileName for lu in fsinfo_local.LUs]
                        ws_lunamelist = [lu.deviceFileName for lu in fsinfo_local.WorkSpaceLUs] if 'ActMig' in fsinfo_local.usage else []
                        is_sufficient = True
                    else:
                        fslulist, wslulist, is_sufficient = HDIOperation.FileSystem.pickup_lu_based_on_policy(fsinfo, policy, storage)
                        fs_capacity = sum([lu.size for lu in fslulist])
                        ws_capacity = sum([lu.size for lu in wslulist])
                        fs_lunamelist = [lu.deviceFileName for lu in fslulist]
                        ws_lunamelist = [lu.deviceFileName for lu in wslulist]

                    if fsinfo.fileSystemName in [e.fileSystemName for e in fslist_local]:
                        print_status("File system %s (already created):" % fsinfo.fileSystemName)
                    else:
                        print_status("File system %s:" % fsinfo.fileSystemName)
                    print_status("    Capacity (file system): %s GB" % fs_capacity)
                    print_status("    Capacity (work space): %s GB" % ws_capacity)
                    print_status("    LU (file system): %s" % fs_lunamelist)
                    print_status("    LU (work space): %s" % ws_lunamelist)
                    print_status("    Policy: %s" % policy.to_display_string())
                    if is_sufficient:
                        print_status("    Sufficient: yes")
                    else:
                        print_status_warning("    Sufficient: NO")
                else:
                    # If the file system was already created, display the information
                    if fsinfo.fileSystemName in [e.fileSystemName for e in fslist_local]:
                        fsinfo_local = [e for e in fslist_local if e.fileSystemName == fsinfo.fileSystemName][0]
                        fs_capacity = fsinfo_local.fileSystemSizeInGigaByte
                        ws_capacity = fsinfo_local.workSpaceSizeInGigaByte if 'ActMig' in fsinfo_local.usage else 0
                        is_sufficient = True
                    else:
                        fs_capacity, ws_capacity, is_sufficient = HDIOperation.FileSystem.get_capacity_based_on_policy(fsinfo, policy, storage)

                    if fsinfo.fileSystemName in [e.fileSystemName for e in fslist_local]:
                        print_status("File system %s (already created):" % fsinfo.fileSystemName)
                    else:
                        print_status("File system %s:" % fsinfo.fileSystemName)
                    print_status("    Capacity (file system): %s GB" % fs_capacity)
                    print_status("    Capacity (work space): %s GB" % ws_capacity)
                    print_status("    Policy: %s" % policy.to_display_string())
                    if is_sufficient:
                        print_status("    Sufficient: yes")
                    else:
                        print_status_warning("    Sufficient: NO")

                if not is_sufficient:
                    has_shortage = True
                    print_status_warning("!!! Cannot allocate sufficient storage for %s !!!" % fsinfo.fileSystemName)

        # Display the remaining storage
        print_status("Remaining storage:")
        if model == HDIConfig.Model.CLUSTER:
            freelulist = storage.get_free_lulist()
            print_status("    Free LUs: %s" % [lu.deviceFileName for lu in freelulist])
            print_status("    Total capacity: %s GB" % sum([lu.size for lu in freelulist]))
        else:
            vgfreelist = storage.get_free_capacity()
            for vgname in vgfreelist:
                print_status("    %s: %s GB" % (vgname, vgfreelist[vgname]))

        # Check MQE status for LFT enabled file systems
        is_MQE_unavailable = False
        if not nocheckMQE:
            print_status("Checking MQE status for LFT enabled file systems ...")
            for node in node_dict[confgroup]['self']:
                archive = archive_dict[node]
                for fsinfo in [fsinfo for fsinfo in archive.fslist if fsinfo.is_supported()]:
                    if fsinfo.LargeFileTransferSetting is None or fsinfo.LargeFileTransferSetting.largeFileTransfer == 'Disable':
                        print_status("LFT is not enabled on %s" % fsinfo.fileSystemName)
                    else:
                        print_status("LFT is enabled on %s" % fsinfo.fileSystemName)
                        namespace = fsinfo.NamespaceShareSettings.hcpNamespace
                        try:
                            if is_search_and_indexing_enabled(namespace, hcp_info, config):
                                print_status("Search and indexing are enabled on %s." % namespace)
                            else:
                                print_status("Search or indexing may not be enabled on %s. Trying to enable ..." % namespace)
                                enable_search_and_indexing(namespace, hcp_info, config)
                        except Exceptions.ConnectionError, e:
                            log_error("Failed to check and enable Seach and Indexing for %s. error=%s. Continue anyway ..." % (namespace, str(e)))
                            is_MQE_unavailable = True
                        except Exceptions.OperationError, e:
                            log_error("Could not enable MQE, Search or indexing, or some error occurred when trying to enable them for %s. reason=%s. Continue anyway ..." % (namespace, str(e)))
                            is_MQE_unavailable = True

    print_status("=== Resource check completed ===")

    return {'has_conflict': has_conflict, 'has_shortage': has_shortage, 'is_MQE_unavailable': is_MQE_unavailable}

def check_conflicts(name_list):
    conflict_name_list = sorted([name for name in set(name_list) if name_list.count(name) > 1])
    return conflict_name_list

def is_search_and_indexing_enabled(namespace, hcp_info, config):
    """ Check if 'Search' and 'Indexing' are enabled on the HCP and namespace.

    This function returns True if Search and Indexing are enabled on the HCP
    and the namespace. Otherwise returns False.

    This function checks Search and Indexing status by issuiung MQE query.
    The query is an object-based query and tries to get a list of objects
    on /system/ on the specified namespace.
    There has to be some objects in that path, so if this function can get
    multiple object information, we can say Search and Indexing are both
    enabled. If this function got 0 results, Search is enabled but Indexing
    may not be enabled (may be enabled but not yet indexed).
    If this function got 503 error, Search or Indexing may not be enabled on
    the HCP or the namespace.

    Args:
        namespace (str): Name of the namespace to check the status.
        hcp_info (HCPClient.HCPInfo): HCP information.
        config (Config.Config): Configuration of the scripts.

    Returns:
        bool: True if Search and Indexing are enabled. False otherwise.
    """

    print_status("Executing test MQE query for %s ..." % namespace)
    # Setup MQE query for checking status
    hcp_client = HCPClient.HCPClient(hcp_info)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)
    hcp_client.enable_content_header()
    query = HCPMQE.HCPMQEObjectQuery()
    query.set_HCPClient(hcp_client)
    query.set_count(10)
    queryexp = "namespace:%s.%s AND objectPath:/system/" % (namespace, hcp_info.tenant)
    query.set_query(queryexp)
    querystr = query.build()
    log_debug("MQE Query: %s" % querystr)

    # Issue query
    try:
        resultset = query.issue()
    except HCPMQE.ConnectionError, e:
        raise Exceptions.ConnectionError(e.reason, e.code)
    except HCPMQE.MQEError, e:
        if e.code == 403:
            # 403 is returned when 'Search' is not enabled on the tenant.
            print_status_warning("403 was returned for MQE query. (detail: %s)" % str(e))
            print_status_warning("Please check if the tenant has 'Search' feature enabled.")
            raise Exceptions.OperationError("403 was returned for MQE query test.", 403)
        elif e.code == 503:
            # 503 is returned when MQE is not enabled on system level.
            print_status_warning("503 was returned for MQE query. (detail: %s)" % str(e))
            print_status_warning("Please check if 'metadata query API' is enabled on HCP system level setting.")
            raise Exceptions.OperationError("503 was returned for MQE query test.", 503)
        else:
            log_error("An unexpected error occurred while executing MQE query. error=%s" % str(e))
            raise Exceptions.OperationError("MQE failure. error=%s" % e.reason, e.code)

    # Count result
    objects = resultset.findall('object')
    if len(objects) == 0:
        print_status_warning("0 objects were returned from HCP. Indexing may not be enabled.")
        result = False
    else:
        print_status("%s objects were returned from HCP." % len(objects))
        result = True

    return result

def enable_search_and_indexing(namespace, hcp_info, config):
    """ Enable Search and Indexing on the namespace.

    This function enables Search and Indexing on the specified namespace by
    issuing MAPI.

    Args:
        namespace (str): Name of namespace to enable Search and Indexing.
        hcp_info (HCPClient.HCPInfo): HCP information.
        config (Config.Config): Configuration of the scripts.

    Returns:
        None
    """

    # Setup HCPClient for MAPI operation
    hcp_client = HCPClient.HCPClient(hcp_info, port=9090)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)
    hcp_client.enable_content_header()

    # Enable Search and index
    path = "/mapi/tenants/%s/namespaces/%s" % (hcp_info.tenant, namespace)
    xml = "<namespace><indexingEnabled>true</indexingEnabled><indexingDefault>true</indexingDefault><searchEnabled>true</searchEnabled></namespace>"

    print_status("Enabling Search and Indexing on %s ..." % namespace)
    try:
        ret = hcp_client.post(path, data=xml)
    except HCPClient.ConnectionError, e:
        log_error("Connection error on issuing MAPI to HCP. error=%s" % str(e))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if ret['code'] != 200:
        if 'X-HCP-ErrorMessage' in ret['headers']:
            errmsg = ret['headers']['X-HCP-ErrorMessage']
        else:
            errmsg = "None"
        if ret['code'] == 403:
            print_status_warning("403 was returned when enabling 'Search' for a namespace. (code=%s, message=%s, error=%s, body=%s)" % (ret['code'], ret['message'], errmsg, ret['body']))
            print_status_warning("Please check if the tenant has 'Search' feature enabled.")
        else:
            log_error("Unexpected HTTP status code. code=%s, message=%s, error=%s, body=%s" % (ret['code'], ret['message'], errmsg, ret['body']))
        print_status_error("Failed to enable Search and Indexing for %s." % namespace)
        raise Exceptions.OperationError("Failed to enable Search and Indexing for %s." % namespace, ret['code'])

    # Start reindexing
    path = "/mapi/tenants/%s/namespaces/%s" % (hcp_info.tenant, namespace)
    xml = "<namespace></namespace>"
    params = {'resetMQECheckpoint': 0}
    print_status("Start reindexing on %s ..." % namespace)
    try:
        hcp_client.set_params(params)
        ret = hcp_client.post(path, data=xml)
    except HCPClient.ConnectionError, e:
        log_error("Connection error on issuing MAPI to HCP. error=%s" % str(e))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if ret['code'] != 200:
        if 'X-HCP-ErrorMessage' in ret['headers']:
            errmsg = ret['headers']['X-HCP-ErrorMessage']
        else:
            errmsg = "None"
        log_error("Unexpected HTTP status code. code=%s, message=%s, error=%s, body=%s" % (ret['code'], ret['message'], errmsg, ret['body']))
        print_status_error("Failed to start reindexing for %s." % namespace)
        raise Exceptions.OperationError("Failed to start reindexing for %s." % namespace, ret['code'])

def get_storage_policy(fsname, target, source, confgroup, config):
    """ Get a storage policy for a file system for a pair with searching
    from all pair configuration.

    This function searches a pair where 'source' is primary and 'target' is
    secondary, from all pair configurations whose configuration group is
    specified by confgroup.
    Once the pair and configuration group are determined, this function
    retrieves a storage policy for a file system specified by 'fsname' from
    the storage policy configuration of the group, and returns the policy.

    If the specified pair is not found in any pair configurations or no
    storage policy is retrieved for the file system, None is returned.

    Args:
        fsname (str): File system name that you want to get a storage
            policy for.
        target (str): IP address of ther target (secondary) node.
        source (str): IP address of the source (primary) node where the
            file system comes from.
        confgroup (str): Configuration group name where the pair is defined.
        config (Config.Config): Configuration of the scripts.
    Returns:
        FailoverConfig.StoragePolicy: Storage policy for the file system.
            None can be returned if the specified pair is not found or
            no storage policy for the file system is found.
    """
    # Load pair configuration
    pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
    stconf = FailoverConfig.StoragePolicyConfig(config.fo_conf_dir, confgroup)

    # Find a pair where primary is 'source' and secondary is 'target'
    my_pairs = pairconf.get_list(primary=source, secondary=target)
    if len(my_pairs) == 0:
        # No such pair is defined in any pair configuration
        log_error("No such pair is defined in any pair configuration. primary=%s, secondary=%s" % (source, target))
        return None
    pid, pair = my_pairs[0]
    policy = stconf.get_policy(pid, fsname)
    if policy is None:
        log_error("Could not get any storage policy applicable to %s. pair_id=%s, pair=%s" % (fsname, pid, pair))

    return policy

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    try:
        short_opts = "h"
        long_opts = ["help", "no-backup", "no-check-MQE"]
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        raise Exceptions.OptionError(e)

    nobackup = False
    nocheckMQE = False
    for o, a in opts:
        if o in ("-h", "--help"):
            raise Exceptions.HelpOption()
        if o == "--no-backup":
            nobackup = True
        if o == "--no-check-MQE":
            nocheckMQE = True

    # Load the existing pair configuration
    log_info("Getting list of pair configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Get my IP address
    myself = setup_common.get_my_ip_address(config)

    # Am I secondary node?
    is_primary = setup_common.is_primary(myself, config.fo_conf_dir)

    if is_primary:
        # I am a primary node.
        # Execute check_resources on the secondary node(s).
        print_status("I am a primary node.")

        for confgroup in confgroup_list:
            # Get my secondary nodes
            pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
            my_pairs = pairconf.get_list(primary=myself)
            print_status("Secondary nodes of my pairs: %s" % [pair.secondary for pid, pair in my_pairs])

            # Setup SSHConfig for this configuration group
            ssh_config = Config.SSHConfig(config.sys_conf_dir, pairconf, confgroup)

            # Setup operation
            options = "--no-backup" if nobackup else ""
            options += " --no-check-MQE" if nocheckMQE else ""
            cmd = "sudo maintexec %s %s" % (os.path.join(BASEDIR, PROGRAM), options)
            op = HDIOperation.CommandOperation(cmd, display_stdout=True)

            # Execute check_resources on each secondary node of my pairs
            for pid, pair in my_pairs:
                cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(pair.secondary), None)
                conn = HDIOperation.Connection(pair.secondary, cred, config.retry_count, config.retry_interval, config.timeout)
                print_status("Executing %s on %s ..." % (PROGRAM, pair.secondary))
                try:
                    ret, out, err = op.execute(conn)
                except HDIOperation.ConnectionError, e:
                    log_error("Failed to connect %s over SSH. code=%s, reason=%s" % (conn.host, e.code, e.reason))
                    raise Exceptions.ConnectionError(e.reason, e.code)

                if ret != 0:
                    log_error("An error occurred while executing %s on %s. ret=%s" % (PROGRAM, pair.secondary, ret))
                    raise Exceptions.OperationError("An error occurred while executing %s on %s." % (PROGRAM, pair.secondary))
    else:
        # If I am a secondary node.
        # Execute check_resources on myself.
        check_resources(myself, config, nobackup, nocheckMQE)

if __name__ == '__main__':
    try:
        main()
    except Exceptions.HelpOption:
        usage()
        sys.exit(0)
    except Exceptions.NoExecution:
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
