#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "failover_filesystem"

import sys
import os
import getopt
import time

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_warning, print_status_error
import HDIConfig
import HDIClient
import HDIOperation
import FailoverConfig

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import Alert
import HDIConfigArchive
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import setup_common
import backup_sync_common
import sync_fs
import failover_common
import getinfo_common

# Constants

TIME_FILESYSTEM = 60
TIME_HISTORY = 5
TIME_ARCRESTORE = 30
TIME_CIFSSHARE = 10
TIME_NFSSHARE = 10
TIME_LFTSETTINGS = 3
TIME_MIGTASK = 15

# Global variable to save information
myself_saved = None
fs_global = None

def usage():
    print "Usage:"
    print "    sudo maintexec %s --source <IP address of source node>" % PROGRAM
    print "                      --file-system <file system name>"
    print "                      [--storage-policy <storage policy string>]"
    print "    sudo maintexec %s -h|--help" % PROGRAM

def get_myself(config):
    global myself_saved
    if myself_saved is None:
        myself_saved = setup_common.get_my_ip_address(config)
    return myself_saved

def find_storage_policy(source, fs, config):
    """ Find a storage policy applied to the file system.

    This function looks for a storage policy that is applied to the file system
    from storage policy configuration.
    This function finds the corresponding pair information from all of
    configuration groups at first, and retrieves storage policy for the
    file system.

    Args:
        source (str): IP address of the source node of failover.
        fs (str): File system name to be failed over.
        config (Config.Config): Configuration of the scripts.
    Returns:
        str: Storage policy string.
    """

    # Get IP address of myself that is used in pair configuration
    myself = get_myself(config)

    # Find a pair whose primary is 'source' and secondary is myself
    confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    for confgroup in confgroup_list:
        pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
        my_pair_list = pairconf.get_list(primary=source, secondary=myself)
        # Found the pair?
        if len(my_pair_list) != 0:
            # Found. Get storage policy for the file system
            pid, notused = my_pair_list[0]
            stconf = FailoverConfig.StoragePolicyConfig(config.fo_conf_dir, confgroup)
            policy = stconf.get_policy(pid, fs)
            break
    else:
        log_error("No storage policy was found for %s from %s. No pair for this failover (primary: %s, secondary: %s) is not defined." % (fs, source, source, myself))
        return None

    return policy.to_string()

def failover_filesystem(source, fs, policy_string, config, rehearsal=None):
    """ Execute failover of a file system.

    This function executes the following processes:
        Download configuration archive of the source node if it doesn't exist.
        Synchronize the file system.
        Restore data from HCP.
        Restore share(s) on the file system.
        Restore migration task(s) on the file system.

    Args:
        source (str): IP address of the failover source node.
        fs (str): File system name to be failed over.
        policy_string (str): String expression of the storage policy to be
            applied to the file system.
            None can be sepcified when pre-configured policy is used.
        config (Config.Config): Configuration of the scripts.
        rehearsal (RehearsalMode): Rehearsal mode specified when rehearsal.
    Returns:
        None
    """

    start = int(time.time())
    print_status("=== Failover of %s from %s started. ===" % (fs, source))
    log_info("failover source node: %s, file system: %s, storage policy: %s, rehearsal: %s" % (source, fs, policy_string, rehearsal))

    # Find a storage policy for the file system if not specified in the arguments
    if policy_string is None:
        policy_string = find_storage_policy(source, fs, config)
        if policy_string is None:
            # No policy found, even default policy cannot be found.
            log_error("Could not get a storage policy for %s. Cannot register this job." % fs)
            print_status_error("Failed to register a failover job for %s due to no storage policy found." % fs)
            raise Exceptions.ConfigurationError("No storage policies for file system %s found." % fs)
        log_info("Storage policy for %s: %s" % (fs, policy_string))

    ### Preparation
    # Setup connection to localhost
    # No SSH credential is required as this is an operation to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Load HCP information
    print_status("Loading HCP information ...")
    try:
        hcpop = HDIOperation.HCPSetting(conn)
        hcp_info = hcpop.get()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HCP information on this node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get HCP configuration on this node. error=%s" % e.reason, e.code)

    # Download configuration archive file of the source node if it doesn't exist
    archive_path = backup_sync_common.get_existing_archive_path(source, config.hdi_conf_dir)
    if archive_path is None:
        # Archive file of the source node does not exist. Let's download it.
        print_status("Downloading configuration archive of %s from HCP ..." % source)
        archive_path = backup_sync_common.download_configs(source, config.hdi_conf_dir, hcp_info, config)
    else:
        log_info("Configuration archive of %s already exists." % source)

    # Load configuration archive
    archive = HDIConfigArchive.HDIConfigArchive()
    archive.load(archive_path)

    # Record result of Preparation
    end = int(time.time())
    failover_common.record_result_preparation(fs, start, end, config.fo_conf_dir)

    ### Synchronize the file system
    print_status("--- File system restoration started. ---")
    # Get file system information from the archive
    fs_list = [fsinfo for fsinfo in archive.fslist if fsinfo.fileSystemName == fs]
    if len(fs_list) == 0:
        log_error("File system %s is not defined on the failover source node %s." % (fs, source))
        raise Exceptions.ConfigurationError("File system %s in not defined on the failover source node %s." % (fs, source))
    fsinfo = fs_list[0]

    # Check if this file system is supported
    if not fsinfo.is_supported():
        print_status("Skip %s because it is not supported." % fsinfo.fileSystemName)
        return

    # Get storage information if this is dry-run failover
    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        # Load storage operation instance from pickle if exists.
        storage = failover_common.load_storage(config.fo_conf_dir)

        # Get storage operation if it was not saved yet.
        if storage is None:
            storage = HDIOperation.Storage(conn)
    else:
        storage = None

    restore_filesystem(fsinfo, source, archive, policy_string, config, rehearsal, storage)

    # Save storage information if this is dry-run failover
    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        failover_common.save_storage(storage, config.fo_conf_dir)

    print_status("--- File system restoration finished. ---")

    ### Restore data from HCP.
    print_status("--- Data restoration from HCP started. ---")
    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        # Load history directory count for dry-run failover
        if fsinfo.fileSystemName in archive.history_count.keys():
            history_count_for_fs = archive.history_count[fsinfo.fileSystemName]
        else:
            print_status_warning("(Dry-run) Number of history directories on %s was not acquired on the source node. Assume 0." % fsinfo.fileSystemName)
            history_count_for_fs = 0
    else:
        history_count_for_fs = None
    restore_data_on_filesystem(fsinfo, config, rehearsal, history_count_for_fs)
    print_status("--- Data restoration from HCP finished. ---")

    ### Restore share(s)
    print_status("--- CIFS share restoration started. ---")
    restore_cifs_share(fsinfo, archive, config, rehearsal)
    print_status("--- CIFS share restoration finished. ---")

    print_status("--- NFS share restoration started. ---")
    restore_nfs_share(fsinfo, archive, config, rehearsal)
    print_status("--- NFS share restoration finished. ---")

    ### Restore Large File Transfer settings
    print_status("--- Large File Transfer settings restoration started (if applicable). ---")
    restore_lft_settings(fsinfo, config, rehearsal)
    print_status("--- Large File Transfer settings restoration finished. ---")

    ### Restore migration task(s)
    print_status("--- Migration task restoration started. ---")
    restore_migration_tasks(fsinfo, archive, config, rehearsal)
    print_status("--- Migration task restoration finished. ---")

    print_status("=== Failover of %s from %s finished. ===" % (fs, source))

def restore_filesystem(fsinfo, source, archive, policy_string, config, rehearsal=None, storage=None):
    """ Restore a file system.

    This function restores a file system specified by fsinfo if not yet, by
    using sync_fs#sync_filesystem() function.

    If this is failback, this function gets information of the original file
    system from configuration archive provided as archive.

    Args:
        fsinfo (HDIConfig.FileSystemInfo): File system information.
        source (str): IP address of the failover source node.
        archive (HDIConfigArchive.HDIConfigArchive): Configuration archive of
            the source node.
        policy_string (str): String format of the storage policy for the file
            system.
        config (Config.Config): Configuration of the scripts.
        rehearsal (RehearsalMode): Rehearsal mode if this is rehearsal.
        storage (HDIOperation.Storage): Storage opration. Required in dry-run
            case.

    Returns:
        None
    """
    start = int(time.time())

    # Setup connection to localhost
    # No SSH credential is required as this is an operation to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Get storage policy
    policy = FailoverConfig.StoragePolicy(fsinfo.fileSystemName, policy_string=policy_string)

    # Get a list of file systems on this node because HEAD against FileSystems resource
    # failes with 503 if no file systems exist...
    fsop = HDIOperation.FileSystem(conn)
    print_status("Getting file systems on this node ...")
    try:
        fslist_on_this_node = fsop.get_all()
    except HDIOperation.OperationError, e:
        log_error("Failed to get a list of file systems on %s. error=%s" % (conn.host, str(e)))
        raise Exceptions.OperationError("Failed to get file system information. error=%s" % e.reason, e.code)

    # Get original file systme information if the policy is 'Failback'
    if policy.policy == FailoverConfig.StoragePolicy.FAILBACK:
        print_status("Policy is 'Failback'. Getting original file system information ...")
        # Get IP address of myself that is used in pair configuration
        myself = get_myself(config)
        if myself not in archive.synced_fslist.keys() or fsinfo.fileSystemName not in archive.synced_fslist[myself].keys():
            log_error("Could not find original file system information in the configuration archive. myself=%s, fs=%s" % (myself, fsinfo.fileSystemName))
            raise Exceptions.ConfigurationError("Original file system information does not exist for failback.")
        original_fsinfo = archive.synced_fslist[myself][fsinfo.fileSystemName]
    else:
        original_fsinfo = None

    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        # Calculate storage configuration
        if storage is None:
            raise Exceptions.ParameterError("Storage operation instance is required for dry-run failover.")
        now = int(time.time())
        if fsinfo.fileSystemName in [e.fileSystemName for e in fslist_on_this_node]:
            # Already exists. Assume we don't need to recreate the file system.
            print_status("(Dry-run) File system %s already exists. Assume we do not need to recreate it." % fsinfo.fileSystemName)
            estimated_time = 1
        else:
            # Calculate storage assignment and time
            model = getinfo_common.get_local_model(config)
            if model == HDIConfig.Model.CLUSTER:
                fslulist, wslulist, is_sufficient = HDIOperation.FileSystem.pickup_lu_based_on_policy(fsinfo, policy, storage, original_fsinfo)
                fs_capacity = sum([lu.size for lu in fslulist])
                ws_capacity = sum([lu.size for lu in wslulist])
                fs_lunamelist = [lu.deviceFileName for lu in fslulist]
                ws_lunamelist = [lu.deviceFileName for lu in wslulist]
            else:
                fs_capacity, ws_capacity, is_sufficient = HDIOperation.FileSystem.get_capacity_based_on_policy(fsinfo, policy, storage, original_fsinfo)
                fs_lunamelist = ws_lunamelist = None

            print_status("(Dry-run) Expected storage configuration for %s:" % fsinfo.fileSystemName)
            print_status("(Dry-run)     Capacity (file system): %s GB" % fs_capacity)
            print_status("(Dry-run)     Capacity (work space): %s GB" % ws_capacity)
            if fs_lunamelist is not None:
                print_status("(Dry-run)     LU (file system): %s" % fs_lunamelist)
            if ws_lunamelist is not None:
                print_status("(Dry-run)     LU (work space): %s" % ws_lunamelist)
            print_status("(Dry-run)     Policy: %s" % policy.to_display_string())
            if is_sufficient:
                print_status("(Dry-run)     Sufficient: yes")
            else:
                print_status_warning("(Dry-run)     Sufficient: NO")

            estimated_time = TIME_FILESYSTEM

        # Calculate estimated time.
        # Simply assume we need 30 seconds to create a file system.
        end = now + estimated_time
    else:
        # Synchronize the file system
        sync_fs.sync_filesystem(fsinfo, source, policy, config, fslist_on_this_node, original_fsinfo)
        end = int(time.time())

    # Record result of file system restoration
    failover_common.record_result_filesystem(fsinfo.fileSystemName, start, end, config.fo_conf_dir)

def restore_data_on_filesystem(fsinfo, config, rehearsal=None, history_count_for_fs=None):
    """ Restore data on a file system from HCP.

    This function executes arcrestore to restore data on the specified file system.
    If this is rehearsal failover, this function executes on-demand restoration.
    Otherwise uses background restoration.

    Args:
        fsinfo (HDIConfig.FileSystemInfo): File system information.
        config (Config.Config): Configuration of the scripts.
        rehearsal (RehearsalMode): Rehearsal mode if this is rehearsal failover.
        history_count_for_fs (int): Number of history directories on the file system on
            the source node. Required if this is dry-run failover.
    Returns:
        None
    """
    start = int(time.time())

    # Setup connection
    # No SSH credential is required as this is an operation to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Namespace information
    namespace = fsinfo.NamespaceShareSettings.hcpNamespace
    log_info("Namespace: %s" % namespace)

    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        # Calculate time to take restore a file system
        now = int(time.time())

        if history_count_for_fs is None:
            raise Exceptions.ParameterError("History count information is required for dry-run failover.")

        # Assumes it takes 5 seconds to restore one history directory
        estimated_time = history_count_for_fs * TIME_HISTORY
        print_status("(Dry-run) %s history directories exist on the file system on the source node." % history_count_for_fs)
        print_status("(Dry-run) %s seconds will required to restore the history directories." % estimated_time)

        # Additional 10 seconds for other tasks
        estimated_time += TIME_ARCRESTORE

        end = now + estimated_time
    else:
        # Restoration mode
        if rehearsal == failover_common.RehearsalMode.TRYOUT:
            mode = ""
        else:
            mode = "--background --fast"

        # Execute arcrestore
        cmd = "sudo arcrestore -f %s --namespace %s --file-system %s" % (mode, namespace, fsinfo.fileSystemName)
        log_info("Executing arcrestore ... (%s)" % cmd)
        print_status("Restoring data on %s ..." % fsinfo.fileSystemName)
        op = HDIOperation.CommandOperation(cmd)
        ret, out, err = op.execute(conn)
        if ret == 80:
            print_status_warning("Data restoration for %s (arcrestore) succeeded, but some history folders might not be restored. ret=%s, out=%s, err=%s" % (fsinfo.fileSystemName, ret, out, err))
        elif ret == 8:
            print_status_warning("Data restoration for %s (arcrestore) was not executed because some data already exists or the file system is not mounted as R/W. Continue anyway ..." % fsinfo.fileSystemName)
        elif ret != 0:
            print_status_error("Data restoration for %s (arcrestore) failed. ret=%s, out=%s, err=%s" % (fsinfo.fileSystemName, ret, out, err))
            raise Exceptions.OperationError("arcrestore failed. error=%s" % err, ret)
        else:
            log_info("Data restoration for %s (arcrestore) succeeded." % fsinfo.fileSystemName)
            print_status("Data restoration succeeded.")
        end = int(time.time())

    # Record result of data restoration
    failover_common.record_result_data(fsinfo.fileSystemName, start, end, config.fo_conf_dir)

def restore_cifs_share(fsinfo, archive, config, rehearsal=None):
    """ Restore CIFS shares on a file system.

    This function uses cifsrestore command to restore CIFS shares on a file system.

    Because cifsrestore requires an XML file that was generated by cifsbackup and
    the file to be on /home/nasroot, this function extracts the XML file from
    configuration archive and put it on /home/nasroot. This is done through
    extract_cifs_xml_file() method of HDIConfigArchive instance.

    cifsrestore is executed through CIFSShare#restore operation.

    Args:
        fsinfo (HDIConfig.FileSystemInfo): Information of the file system where
            CIFS shares are restored on.
        archive (HDIConfigArchive.HDIConfigArchive): HDI configuration archive
            instance.
        config (Config.Config): Configuration of the scripts.
        rehearsal (RehearsalMode): Rehearsal mode if this is rehearsal.
    Returns:
        None
    """
    start = int(time.time())

    # Setup connection
    # No SSH credential is required as this is an operation to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Extract CIFS xml file for the file system
    print_status("Extracting CIFS configuration file for %s from archive ..." % fsinfo.fileSystemName)
    xml_path = archive.extract_cifs_xml_file(fsinfo.fileSystemName, '/home/nasroot')
    if xml_path is None:
        log_info("CIFS backup file for file system %s does not exist." % fsinfo.fileSystemName)
        print_status("No CIFS shares for %s." % fsinfo.fileSystemName)
        return

    # Count CIFS shares for recording purpose
    cifslist = HDIConfig.parse_cifsshares(xml_file=xml_path)
    num = len(cifslist)

    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        # Calculate time to restore the shares
        # Assumes it will take a fixed time for each share.
        now = int(time.time())
        print_status("(Dry-run) There are %s CIFS shares." % num)
        print_status("(Dry-run) CIFS shares: %s" % [cifsinfo.cifsShareName for cifsinfo in cifslist])
        estimated_time = num * TIME_CIFSSHARE
        end = now + estimated_time
    else:
        # Execute CIFS restoration
        cifsop = HDIOperation.CIFSShare(conn)
        print_status("Restoring CIFS shares on %s ..." % fsinfo.fileSystemName)
        try:
            cifsop.restore(xml_path)
        except HDIOperation.OperationError, e:
            if e.reason.find('KAQM08150-E') != -1:
                # CIFS shares already exist on the file system. Continue to proceed.
                print_status_warning("File system %s already has a CIFS share. Continue the process ..." % fsinfo.fileSystemName)
            else:
                # Stop process for other errors.
                print_status_error("Failed to restore CIFS shares on %s. error=%s" % (fsinfo.fileSystemName, str(e)))
                raise Exceptions.OperationError("Failed to restore CIFS shares on %s. error=%s" % (fsinfo.fileSystemName, e.reason), e.code)

        Utility.remove_file(xml_path)
        print_status("CIFS share restoration on %s succeeded." % fsinfo.fileSystemName)
        end = int(time.time())

    # Record result of CIFS share restoration
    failover_common.record_result_cifs(fsinfo.fileSystemName, num, start, end, config.fo_conf_dir)

def restore_nfs_share(fsinfo, archive, config, rehearsal=None):
    """ Restore NFS shares on a file system.

    Same as restore_cifs_share() other than the fact that this function restores
    NFS shares.

    Args:
        fsinfo (HDIConfig.FileSystemInfo): Information of the file system where
            NFS shares are restored on.
        archive (HDIConfigArchive.HDIConfigArchive): HDI configuration archive
            instance.
        config (Config.Config): Configuration of the scripts.
        rehearsal (RehearsalMode): Rehearsal mode if this is rehearsal.
    Returns:
        None
    """
    start = int(time.time())

    # Setup connection
    # No SSH credential is required as this is an operation to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Extract NFS xml file for the file system
    print_status("Extracting NFS configuration file for %s from archive ..." % fsinfo.fileSystemName)
    xml_path = archive.extract_nfs_xml_file(fsinfo.fileSystemName, '/home/nasroot')
    if xml_path is None:
        log_info("NFS backup file for file system %s does not exist." % fsinfo.fileSystemName)
        print_status("No NFS shares for %s." % fsinfo.fileSystemName)
        return

    # Count NFS shares for recording purpose
    nfslist = HDIConfig.parse_nfsshares(xml_file=xml_path)
    num = len(nfslist)

    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        # Calculate time to restore the shares
        # Assumes it will take a fixed time for each share.
        now = int(time.time())
        print_status("(Dry-run) There are %s NFS shares." % num)
        print_status("(Dry-run) NFS shares: %s" % [nfsinfo.exportPoint for nfsinfo in nfslist])
        estimated_time = num * TIME_NFSSHARE
        end = now + estimated_time
    else:
        # Execute NFS restoration
        nfsop = HDIOperation.NFSShare(conn)
        print_status("Restoring NFS shares on %s ..." % fsinfo.fileSystemName)
        try:
            nfsop.restore(xml_path)
        except HDIOperation.OperationError, e:
            if e.reason.find('KAQM08150-E') != -1:
                # NFS shares already exist on the file system. Continue to proceed.
                print_status_warning("File system %s already has a NFS share. Continue the process ..." % fsinfo.fileSystemName)
            else:
                # Stop process for other errors.
                print_status_error("Failed to restore NFS shares on %s. error=%s" % (fsinfo.fileSystemName, str(e)))
                raise Exceptions.OperationError("Failed to restore NFS shares on %s. error=%s" % (fsinfo.fileSystemName, e.reason), e.code)

        Utility.remove_file(xml_path)
        print_status("NFS share restoration on %s succeeded." % fsinfo.fileSystemName)
        end = int(time.time())

    # Record result of NFS share restoration
    failover_common.record_result_nfs(fsinfo.fileSystemName, num, start, end, config.fo_conf_dir)

def restore_lft_settings(fsinfo, config, rehearsal=None):
    """ Restore Large File Transfer settings on a file system.

    This function sets parameters of Large File Transfer function defined in
    fsinfo (FileSystemInfo) for the file system.
    If this node (localhost) does not support LFT, nothing is done.
    HDIOperation#FileSystem#set_lft() checks the version of the node, and
    returns with doing nothing if the version is less than the version LFT
    is supported.

    Args:
        fsinfo (HDIConfig.FileSystemInfo): File system information.
        config (Config.Config): Configuration of the scripts.
        rehearsal (RehearsalMode): Rehearsal mode if this is rehearsal.

    Returns:
        None
    """
    start = int(time.time())

    # Setup connection
    # No SSH credential is required as this is an operation to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        # Calculate time to restore LFT settings
        now = int(time.time())
        estimated_time = TIME_LFTSETTINGS
        end = now + estimated_time
    else:
        # Apply LFT settings
        fsop = HDIOperation.FileSystem(conn)
        print_status("Restoring Large File Transfer settings on %s ..." % fsinfo.fileSystemName)
        try:
            fsop.set_lft(fsinfo.fileSystemName, fsinfo.LargeFileTransferSetting)
        except HDIOperation.OperationError, e:
            if e.code == 80:
                print_status_warning("Setting Large File Transfer finished but there might be something to be checked. message=%s" % e.reason)
            else:
                print_status_error("Failed to set Large File Transfer settings for %s. error=%s" % (fsinfo.fileSystemName, str(e)))

        else:
            print_status("Large File Transfer settings restoration on %s succeeded." % fsinfo.fileSystemName)
        end = int(time.time())

    # Record result
    failover_common.record_result_other(fsinfo.fileSystemName, start, end, config.fo_conf_dir)

def restore_migration_tasks(fsinfo, archive, config, rehearsal=None):
    """ Restore migration tasks on a file system.

    This function restores migration tasks on a file system.
    Task information is acquired from HDI configuration archive instance (archive).
    Execute task restoration through MigrationTask#create operation.

    Args:
        fsinfo (HDIConfig.FileSystemInfo): Information of the file system where
            migration tasks are restored on.
        archive (HDIConfigArchive.HDIConfigArchive): HDI configuration archive
            instance.
        config (Config.Config): Configuration of the scripts.
        rehearsal (RehearsalMode): Rehearsal mode if this is rehearsal.
    Returns:
        None
    """
    start = int(time.time())

    # Setup connection
    # No SSH credential is required as this is an operation to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Get migration task information for this file system
    tasklist = [taskinfo for taskinfo in archive.tasklist if taskinfo.fileSystemName == fsinfo.fileSystemName]
    if len(tasklist) == 0:
        print_status_warning("No tasks are defined for file system %s." % fsinfo.fileSystemName)
        return
    log_info("Migration tasks for %s: %s" % (fsinfo.fileSystemName, [t.task_name for t in tasklist]))

    num = len(tasklist)

    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        # Calculate time to restore all tasks
        now = int(time.time())
        estimated_time = num * TIME_MIGTASK

        # Just display task start time for demonstration
        for taskinfo in tasklist:
            now2 = int(time.time())
            start_time = now2 + 86400
            print_status("(Dry-run) Task %s will start at %s." % (taskinfo.task_name, Utility.epoch_to_iso8601(start_time)))

        end = now + estimated_time
    else:
        # Restore each task
        taskop = HDIOperation.MigrationTask(conn)
        for taskinfo in tasklist:
            print_status("Restoring task %s on %s ..." % (taskinfo.task_name, fsinfo.fileSystemName))

            # Check existence of the task
            try:
                exists = taskop.exists(taskinfo.task_name, taskinfo.fileSystemName)
            except HDIOperation.OperationError, e:
                print_status_error("Failed to check existence of migration task %s. Skip this task... error=%s" % (taskinfo.task_name, str(e)))
                continue

            if exists:
                print_status("Task %s already exists. Delete the task to recreate ..." % taskinfo.task_name)
                # Delete the task to recreate.
                # Recreate the task if it exists to apply the new start time.
                try:
                    taskop.delete(taskinfo.task_name, taskinfo.fileSystemName)
                except HDIOperation.OperationError, e:
                    print_status_error("Failed to delete task %s. error=%s. Please re-configure the task manually later ..." % (taskinfo.task_name, str(e)))
                    # Continue anyway to restore remaining tasks...
                    continue

            # Calculate time to start the task
            # Set 24 hours later from now
            now = int(time.time())
            start_time = now + 86400
            print_status("Task %s will start at %s." % (taskinfo.task_name, Utility.epoch_to_iso8601(start_time)))

            # Create the task
            try:
                taskop.create(taskinfo, start_time)
            except HDIOperation.OperationError, e:
                print_status_error("Failed to create migration task %s. error=%s" % (taskinfo.task_name, str(e)))
                # Continue anyway to restore remaining tasks...
                continue

            # Disable the task if this is rehearsal failover or LFT is enabled
            lft = fsinfo.LargeFileTransferSetting is not None and fsinfo.LargeFileTransferSetting.largeFileTransfer == 'Enable'
            if rehearsal == failover_common.RehearsalMode.TRYOUT or lft:
                if rehearsal == failover_common.RehearsalMode.TRYOUT:
                    print_status("Disabling task %s as this is rehearsal failover ..." % taskinfo.task_name)
                elif lft:
                    print_status("Disabling task %s as Large File Transfer is enabled ... The task will be enabled when executing complete_failover or complete_failback." % taskinfo.task_name)
                try:
                    taskop.disable(taskinfo.task_name, taskinfo.fileSystemName)
                except HDIOperation.OperationError, e:
                    log_error("Failed to disable task %s. error=%s" % (taskinfo.task_name, str(e)))
                    # Continue anyway to restore remaining tasks...
                    continue

        print_status("Migration task restoration on %s finished." % fsinfo.fileSystemName)

        end = int(time.time())

    # Record result of task restoration
    failover_common.record_result_task(fsinfo.fileSystemName, num, start, end, config.fo_conf_dir)

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    try:
        short_opts = "h"
        long_opts = ["id=", "source=", "file-system=", "storage-poliy=", "dry-run", "try-out", "help"]
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        raise Exceptions.OptionError(e)

    source = None
    fs = None
    policy_string = None
    rehearsal = None

    for o, a in opts:
        if o == "--id":
            Utility.ID = a
        if o == "--source":
            source = a
        if o == "--file-system":
            fs = a
            global fs_global
            fs_global = fs
        if o == "--storage-policy":
            policy_string = a
        if o == "--dry-run":
            rehearsal = failover_common.RehearsalMode(failover_common.RehearsalMode.DRYRUN)
        if o == "--try-out":
            rehearsal = failover_common.RehearsalMode(failover_common.RehearsalMode.TRYOUT)
        if o in ("-h", "--help"):
            raise Exceptions.HelpOption()

    if source is None or fs is None:
        raise Exceptions.OptionError("Source IP address and file system name are mandatory.")

    # Execute filesystem failover
    failover_filesystem(source, fs, policy_string, config, rehearsal=rehearsal)

if __name__ == '__main__':
    try:
        main()
    except Exceptions.HelpOption:
        usage()
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail="Interrupted by a signal.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail="Unable to load script configuration file.")
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(20)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(99)

    # Send an OK alert if it finished without an error
    Alert.send_alert(Alert.MsgCode.FAILOVER_OK, fsname=fs_global)
    sys.exit(0)
